/*****************************************************************************
 * Copyright © 2011 - 2013, M2X BV
 * Author: Jean-Paul Saman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <assert.h>

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_vout_display.h>
#include <vlc_vout_window.h>
#include <vlc_picture_pool.h>

#include <vlc_arrays.h>

#include <xcb/xcb.h>
#include <X11/Xlib-xcb.h>
#include <vlc_xlib.h>

#include <X11/Xutil.h>
#include "../../video_output/xcb/xcb_events_vlc.h"

#ifdef HAVE_LIBAVCODEC_AVCODEC_H
#   include <libavcodec/avcodec.h>
#   include <libavcodec/vaapi.h>
#else
#   include <avcodec.h>
#   include <vaapi.h>
#endif

#include "../../codec/avcodec/avcodec.h"

#include <va/va.h>
#include <va/va_x11.h>

#include "../../codec/avcodec/va.h"
#include "../../codec/avcodec/vaapi.h"

#include "common.h"
#include "spu.h"

vout_window_t *MakeWindow(vout_display_t *vd)
{
    vout_window_cfg_t wnd_cfg;

    memset (&wnd_cfg, 0, sizeof (wnd_cfg));
    wnd_cfg.type = VOUT_WINDOW_TYPE_XID;
    wnd_cfg.x = var_InheritInteger(vd, "video-x");
    wnd_cfg.y = var_InheritInteger(vd, "video-y");
    wnd_cfg.width  = vd->cfg->display.width;
    wnd_cfg.height = vd->cfg->display.height;

    vout_window_t *wnd = vout_display_NewWindow(vd, &wnd_cfg);
    if (wnd == NULL)
        msg_Err(vd, "parent window not available");
    return wnd;
}

const xcb_screen_t *FindWindow(vout_display_t *vd,
           vout_window_t *embed, xcb_connection_t *conn,
           unsigned *restrict pnum, uint8_t *restrict pdepth,
           uint16_t *restrict pwidth, uint16_t *restrict pheight)
{
    xcb_get_geometry_reply_t *geo =
        xcb_get_geometry_reply(conn,
            xcb_get_geometry(conn, embed->handle.xid), NULL);
    if (geo == NULL)
    {
        msg_Err (vd, "parent window not valid");
        return NULL;
    }

    xcb_window_t root = geo->root;
    *pdepth = geo->depth;
    *pwidth = geo->width;
    *pheight = geo->height;
    free (geo);

    /* Find the selected screen */
    const xcb_setup_t *setup = xcb_get_setup(conn);
    const xcb_screen_t *screen = NULL;
    unsigned num = 0;

    for (xcb_screen_iterator_t i = xcb_setup_roots_iterator(setup);
         i.rem > 0;
         xcb_screen_next (&i))
    {
        if (i.data->root == root)
        {
            screen = i.data;
            break;
        }
        num++;
    }

    if (screen == NULL)
    {
        msg_Err(vd, "parent window screen not found");
        return NULL;
    }
    msg_Dbg(vd, "using screen 0x%"PRIx32 " (number: %u)", root, num);
    *pnum = num;
    return screen;
}

int CreateWindow(vout_display_t *vd,
                 vout_window_t *embed, xcb_connection_t *conn,
                 uint_fast8_t depth, xcb_visualid_t vid,
                 uint_fast16_t width, uint_fast16_t height,
                 xcb_window_t *restrict window)
{
    const uint32_t mask = XCB_CW_EVENT_MASK;
    const uint32_t values[] = {
        /* XCB_CW_EVENT_MASK */
        XCB_EVENT_MASK_VISIBILITY_CHANGE,
    };
    xcb_void_cookie_t cc, cm;

    cc = xcb_create_window_checked(conn, depth, *window,
                                   embed->handle.xid, 0, 0,
                                   width, height, 0,
                                   XCB_WINDOW_CLASS_INPUT_OUTPUT,
                                   vid, mask, values);
    cm = xcb_map_window_checked(conn, *window);
    if (CheckError(vd, conn, "cannot create XCB window", cc) ||
        CheckError(vd, conn, "cannot map XCB window", cm))
        return VLC_EGENERIC;

    msg_Dbg (vd, "using VAAPI XCB window %08"PRIx32, *window);
    xcb_flush(conn);
    return VLC_SUCCESS;
}

/**
 * PictureLock and PictureUnlock will be called by picture_pool_*
 * when the picture is held and released. Both are called only
 * once. PictureLock when the picture is retrieved from the pool
 * and PictureUnlock when the picture is returned to the pool.
 */
static int PictureLock(picture_t *picture)
{
    vlc_va_surface_t *surface = picture->p_sys;
    if (surface == NULL)
        abort();

    vlc_va_SurfaceHold(surface);
    return VLC_SUCCESS;
}

static void PictureUnlock(picture_t *picture)
{
    assert(picture->p_sys);

    vlc_va_surface_t *surface = picture->p_sys;
    if (surface == NULL)
        abort();

    vlc_va_SurfaceRelease(surface);
}

/* PictureRelease is called when the picture pool is destroyed. */
static void PictureRelease(picture_t *picture)
{
    if (picture->i_refcount > 0)
        picture->i_refcount--;

    picture->p_sys = NULL;
}

/**
 * Create a pool of direct buffers to use.
 */
picture_pool_t *vout_vaapi_GetPool(vout_display_t *vd, vlc_va_conn_t *va, unsigned requested)
{
    picture_pool_t *pool = NULL;

    assert(va);
    assert(vd);
    assert(requested > 0);

    /* Create surfaces */
    va->lock();
    unsigned int alloc_count = __MAX(vlc_va_conn_SurfacesCount(va), requested);
    va->unlock();

    picture_t **pic_array = (picture_t **) calloc(alloc_count, sizeof(picture_t *));
    if (!pic_array)
       return NULL;

    unsigned int count;
    for (count = 0; count < alloc_count; count++)
    {
        picture_t *pic = (picture_t*) picture_NewFromFormat(&vd->fmt);
        if (!pic)
            break;
        assert(pic->p_sys == NULL);

        pic->format.i_chroma = VLC_CODEC_VAAPI_SURFACE;

        pic_array[count] = pic;
        pic_array[count]->b_progressive = true;
        pic_array[count]->pf_release = PictureRelease;
    }

    assert( count > 0 );
    if (count == 0)
        goto error;

    unsigned int i_surface_count = __MIN(count, alloc_count);
    assert(i_surface_count == count);

    msg_Info(vd, "Requested %d/%d surfaces for picture pool.",
             i_surface_count, vlc_va_conn_SurfacesCount(va));

    /* Create surfaces */
    va->lock();
    if (!vlc_va_conn_CreateSurface(va,
                                   vd->fmt.i_visible_width,
                                   vd->fmt.i_visible_height,
                                   VA_RT_FORMAT_YUV420, i_surface_count))
    {
        /* FAIL */
        va->unlock();
        goto error;
    }

    for (unsigned int i = 0; i < i_surface_count; i++)
    {
        vlc_va_surface_t *surface = vlc_va_conn_GetSurface(va);
        if( surface )
            pic_array[i]->p_sys = (picture_sys_t *)surface;
    }
    va->unlock();

    /* Register pool with video output core */
    picture_pool_configuration_t cfg;

    memset(&cfg, 0, sizeof(cfg));
    cfg.picture_count = count;
    cfg.picture       = pic_array;
    cfg.lock          = PictureLock;
    cfg.unlock        = PictureUnlock;

    pool = picture_pool_NewExtended(&cfg);
    if (!pool)
        goto error;

    free(pic_array);

    return pool;

error:
    /* Cleanup on error */
    va->lock();
    for (unsigned int i = 0; i < count; i++)
    {
        picture_t *pic = pic_array[i];
        if (!pic)
            continue;

        if (pic->p_sys)
        {
            vlc_va_surface_t *surface = (vlc_va_surface_t *) pic->p_sys;
            vlc_va_conn_ReleaseSurface(va, surface);
        }
        picture_Release(pic);
    }
    free(pic_array);
    va->unlock();

    return NULL;
}
