/*****************************************************************************
 * Copyright © 2012 - 2013, M2X BV
 * Author: Jean-Paul Saman
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_es.h>
#include <vlc_picture.h>
#include <vlc_subpicture.h>

#include <vlc_arrays.h>

#include <xcb/xcb.h>
#include <X11/Xlib-xcb.h>

#include <va/va.h>

#include "../../codec/avcodec/vaapi.h"
#include "spu.h"

/* subpicture region cache management */
static unsigned int cache_counter = 0;

struct vlc_va_spu_cache_t
{
    vlc_mutex_t lock;
    vlc_array_t cache;
};

typedef struct va_spu_cache_t
{
  VASubpictureID i_id;
  VAImage        *image;
  vlc_array_t    vasurfaces; /* list of surfaces associated to */
  struct {
    int i_x, i_y;
    int i_width, i_height;
  } rect;
} va_spu_cache_t;

struct spu_cache_t
{
    unsigned int  i_id;
    unsigned int  i_refcount;  /* refcount usage */
    mtime_t       i_start;     /* start display date */
    mtime_t       i_stop;      /* stop display date */
    vlc_array_t   subpictures; /* subpicture regions */
};

/* */
static spu_cache_t *cache_new(void)
{
    spu_cache_t *cache = calloc(1, sizeof(spu_cache_t));
    if (cache == NULL)
        return NULL;

    /* did we reach the maximum number of subpictures that can be created? */
    if (cache_counter == UINT_MAX)
    {
        abort();
        return NULL;
    }

    cache->i_id = ++cache_counter;
    assert(cache_counter > 0);
    vlc_array_init(&cache->subpictures);
    return cache;
}

static void cache_destroy(spu_cache_t *cache)
{
    assert(cache);
    assert(cache->i_refcount == 0);

    const int count = vlc_array_count(&cache->subpictures);
    if (count != 0)
        abort();

    vlc_array_clear(&cache->subpictures);
    free(cache);
}

static spu_cache_t *cache_find(vlc_array_t *cache, const unsigned int id)
{
    assert(cache);

    const int count = vlc_array_count(cache);
    for (int i = 0; i < count; i++)
    {
        spu_cache_t *c = (spu_cache_t *) vlc_array_item_at_index(cache, i);
        if (!c)
            continue;
        if (c->i_id == id)
            return c;
    }
    return NULL;
}

static inline void cache_ref(spu_cache_t *cache)
{
    assert(cache);
    cache->i_refcount++;
}

static inline void cache_unref(spu_cache_t *cache)
{
    assert(cache);
    assert(cache->i_refcount > 0);
    if (cache->i_refcount > 0)
        cache->i_refcount--;
}

static va_spu_cache_t *subpicture_cache_new( void )
{
    va_spu_cache_t *cache = calloc(1, sizeof(va_spu_cache_t));
    if (cache == NULL)
        return NULL;

    cache->i_id = VA_INVALID_ID;
    cache->image = calloc(1, sizeof(VAImage));
    if (cache->image == NULL)
    {
        free(cache);
        return NULL;
    }
    vlc_array_init(&cache->vasurfaces);
    return cache;
}

static void subpicture_cache_destroy(va_spu_cache_t *cache)
{
    assert(cache);
    assert(cache->i_id == VA_INVALID_ID);
    assert(cache->image);

    const int count = vlc_array_count(&cache->vasurfaces);
#if 0
    if (count != 0)
        abort();
#else
    for (int i = 0; i < count; i++)
    {
        VASurfaceID *va_id = (VASurfaceID *) vlc_array_item_at_index(&cache->vasurfaces, i);
        if (va_id && (*va_id != VA_INVALID_ID))
            abort(); /* BUG: if this happens then there is a bug in the code */
    }
#endif
    vlc_array_clear(&cache->vasurfaces);
    free(cache->image);
    free(cache);
    cache = NULL;
}

static void subpicture_cache_link(va_spu_cache_t *cache, VASurfaceID *id)
{
    assert(cache);

    vlc_array_t *vasurfaces = &cache->vasurfaces;
    const int count = vlc_array_count(vasurfaces);
    for (int i = 0; i < count; i++)
    {
        VASurfaceID *va_id = (VASurfaceID *) vlc_array_item_at_index(vasurfaces, i);
        if (va_id)
        {
            if (*va_id == *id)
                /* NOTE: alread linked to surface */
                abort();
            continue;
        }

        vlc_array_append(vasurfaces, id);
    }
}

static void subpicture_cache_unlink(va_spu_cache_t *cache, VASurfaceID *id)
{
    assert(cache);

    vlc_array_t *vasurfaces = &cache->vasurfaces;
    const int count = vlc_array_count(vasurfaces);
    for (int i = 0; i < count; i++)
    {
        VASurfaceID *va_id = (VASurfaceID *) vlc_array_item_at_index(vasurfaces, i);
        if (!va_id)
        {
            /* NOTE: alread unlinked from surface */
            abort();
        }

        if (*va_id == *id)
            vlc_array_remove(vasurfaces, i);
    }
}

unsigned int vlc_va_spu_cache_GetID(const spu_cache_t *cache)
{
    assert(cache);
    return cache->i_id;
}

/* ------------------------ */
/* VA API support functions */
/* ------------------------ */
int VASubtitleFourCC(vlc_va_conn_t *va, const unsigned int vafourcc, VAImageFormat *sub_fmt, int *flags)
{
    va->lock();

    int num = vaMaxNumSubpictureFormats(va->p_display);
    if (num <= 0)
        goto out_warning;

    VAImageFormat *p_fmt = calloc(num, sizeof(*p_fmt));
    unsigned int  *vaflags = calloc(num, sizeof(*vaflags));
    if (!p_fmt || !vaflags)
        goto out;

    unsigned int num_fmt = 0;
    VAStatus status = vaQuerySubpictureFormats(va->p_display,
                                               p_fmt, vaflags, &num_fmt);
    if (status != VA_STATUS_SUCCESS)
        goto out;

    for (unsigned int i = 0; i < num_fmt; i++)
    {
        VAImageFormat format = p_fmt[i];
        if (format.fourcc == vafourcc)
        {
            *flags = vaflags[i];
            memcpy((void*)sub_fmt, (void*)&format, sizeof(VAImageFormat));
            free(p_fmt);
            free(vaflags);
            va->unlock();
            return VLC_SUCCESS;
        }
    }

out:
    free(p_fmt);
    free(vaflags);

out_warning:
    va->unlock();
    return VLC_EGENERIC;
}

#if !VA_CHECK_VERSION(0,33,0)
static inline void vaapi_fixup_alpha(uint8_t *p_pixel, size_t i_size)
{
    for (size_t p = 0; p < i_size; p += 4)
    {
        /* RGBA */
        uint8_t *R = &p_pixel[p];
        uint8_t *G = &p_pixel[p+1];
        uint8_t *B = &p_pixel[p+2];
        uint8_t *A = &p_pixel[p+3];

        if (*A != 255) /* alpha transparancy */
        {
            *A = 0; /* 0 = transparent, 255 = opaque */
            *R = 0;
            *G = 0;
            *B = 0;
        }
    }
}
#endif

static int CopyRGBAToVAImage(vlc_object_t *obj, VADisplay p_display,
                             picture_t *pic, VAImage *image, VAImageFormat *fmt,
                             const bool fixup_alpha)
{
    assert(image);
    assert(pic->format.i_chroma == VLC_CODEC_RGBA);
#if VA_CHECK_VERSION(0,33,0)
    VLC_UNUSED(fixup_alpha);
#endif

    /* Only supports RGBA images */
    if (!image || !fmt ||
        (pic->format.i_chroma != VLC_CODEC_RGBA))
        return VLC_EGENERIC;

    VAStatus status = vaCreateImage(p_display, fmt, pic->format.i_width,
                                    pic->format.i_height, image);
    if (status != VA_STATUS_SUCCESS)
    {
        msg_Err(obj, "failed to creating image (%d)", status);
        return VLC_EGENERIC;
    }

    /* Sanity checks */
    if ((image->image_id == VA_INVALID_ID) ||
        (image->buf == VA_INVALID_ID) ||
        (image->format.fourcc != VA_FOURCC_RGBA) ||
        (image->num_planes != 1) ||
        (image->height != pic->format.i_height) ||
        (image->width != pic->format.i_width) ||
        (image->format.bits_per_pixel != pic->format.i_bits_per_pixel))
    {
        msg_Err(obj, "image created with wrong properties, destroying it.");
        vaDestroyImage(p_display, image->image_id);
        return VLC_EGENERIC;
    }

    void *p_base;
    status = vaMapBuffer(p_display, image->buf, &p_base);
    if (status != VA_STATUS_SUCCESS)
    {
        msg_Err(obj, "failed mapping image for copying (%d)", status);
        msg_Err(obj, "subpicture region: planes %d/%d, depth %d, width %d/%d, height %d/%d, bpp %d/%d",
                 image->num_planes, pic->i_planes,
                 image->format.depth,
                 image->width, pic->format.i_width,
                 image->height, pic->format.i_height,
                 image->format.bits_per_pixel, pic->format.i_bits_per_pixel);
        vaDestroyImage(p_display, image->image_id);
        return VLC_EGENERIC;
    }
    assert(p_base); /* should not happen */

    /* Does the pitch match? Then just copy it directly */
    if (image->pitches[0] == (unsigned int)pic->p[0].i_pitch)
    {
        size_t i_size = __MIN(image->data_size,
                             (unsigned int)pic->p[0].i_pitch * pic->p[0].i_visible_lines);
        uint8_t *p_dest = (uint8_t*)(((uint8_t*)p_base) + image->offsets[0]);
        uint8_t *p_src  = pic->p[0].p_pixels;

        memcpy(p_dest, p_src, i_size);
    }
    else
    {
        /* proceed line by line */
        size_t bytes_copied = 0;
        uint8_t *p_in  = pic->p[0].p_pixels;
        uint8_t *p_out = (uint8_t *)(((uint8_t*)p_base) + image->offsets[0]);

        for (int i_line = pic->format.i_y_offset * pic->p[0].i_visible_pitch;
                 i_line < pic->p[0].i_visible_lines; i_line++)
        {
            /* NOTE: image->pitch[0] includes image->offset[0], thus
             * visible_pitch = image->pitch[0] - image->offset[0] */
            size_t i_size = __MIN(image->pitches[0],
                                 (unsigned int)pic->p[0].i_pitch);
            uint8_t *p_dest = p_out + (i_line * image->pitches[0]);
            uint8_t *p_src  = p_in + (i_line * pic->p[0].i_pitch);

            memcpy(p_dest, p_src, i_size);

            bytes_copied += i_size;
            assert(bytes_copied <= image->data_size);

#if !VA_CHECK_VERSION(0,33,0)
            /* FIXME: Remove this code when VDPAU and XvBA backends for
             * VAAPI properly support alpha or chroma keying in subpictures. */
            if (fixup_alpha)
            {
                uint8_t *p_pixel = p_out + (i_line * image->pitches[0]);
                vaapi_fixup_alpha(p_pixel, i_size);
            }
#endif
        }
    }

    p_base = NULL;
    status = vaUnmapBuffer(p_display, image->buf);
    if (status != VA_STATUS_SUCCESS)
    {
        msg_Err(obj, "failed unmapping image (%d) for copying (%d)",
                image->image_id, status);
        vaDestroyImage(p_display, image->image_id);
        return VLC_EGENERIC;
    }

    return VLC_SUCCESS;
}

/* Create subpicture cache for this subpicture and
 * allocate needed VA API resources. Call this
 * function when holding cache->lock */
static va_spu_cache_t *va_spu_cache_Create(vlc_object_t *obj, vlc_va_conn_t *va, VAImageFormat *fmt,
            const subpicture_t *subpicture, const subpicture_region_t *region, const int flags)
{
    assert(subpicture);

    va_spu_cache_t *vasub_cache = subpicture_cache_new();
    if (unlikely(!vasub_cache))
        return NULL;

    va->lock();

    /* Copy pixels from region into VAImage */
    VAImage *image = vasub_cache->image;
    int err = CopyRGBAToVAImage(obj, va->p_display, region->p_picture,
                                image, fmt, (flags == 0x0) ? true : false);
    if (err != VLC_SUCCESS)
    {
        subpicture_cache_destroy(vasub_cache);
        va->unlock();
        return NULL;
    }

    VAStatus status;
    status = vaCreateSubpicture(va->p_display, image->image_id, &vasub_cache->i_id);
    if (status != VA_STATUS_SUCCESS)
        goto cleanup;

    assert(vasub_cache->i_id != VA_INVALID_ID);

    bool b_alpha  = (flags & VA_SUBPICTURE_GLOBAL_ALPHA);
    bool b_chroma = (flags & VA_SUBPICTURE_CHROMA_KEYING);

    if (b_alpha)
    {
        assert(!b_chroma);

        float global_alpha = (float)(subpicture->i_alpha * region->i_alpha / 255) / 255;
        status = vaSetSubpictureGlobalAlpha(va->p_display, vasub_cache->i_id, global_alpha);
        if (status != VA_STATUS_SUCCESS)
            msg_Warn(obj, "failed applying alpha value to subpicture");
    }

    if (b_chroma)
    {
        assert(!b_alpha);

        /* chromakey values */
        const unsigned int min = 0x0; /* ? */
        const unsigned int max = 0x0000ff00; /* ? */
        const unsigned int mask = (region->p_picture->format.i_rmask << 24) |
                                  (region->p_picture->format.i_gmask << 16) |
                                  (region->p_picture->format.i_bmask <<  8) | 0;

        status = vaSetSubpictureChromakey(va->p_display, vasub_cache->i_id,
                                          min, max, mask);
        if (status != VA_STATUS_SUCCESS)
            msg_Warn(obj, "failed applying chroma values to subpicture");
    }

    vasub_cache->rect.i_x = region->i_x;
    vasub_cache->rect.i_y = region->i_y;
    vasub_cache->rect.i_width = region->fmt.i_visible_width;
    vasub_cache->rect.i_height = region->fmt.i_visible_height;

    va->unlock();
    return vasub_cache;

cleanup:
    vaDestroyImage(va->p_display, image->image_id);
    vasub_cache->i_id = VA_INVALID_ID;
    subpicture_cache_destroy(vasub_cache);

    va->unlock();
    return NULL;
}

/* Destroy subpicture cache for this subpicture and
 * release allocated VA API resources. Call this
 * function when holding cache->lock */
static void va_spu_cache_Destroy(vlc_va_conn_t *va, spu_cache_t *cache, const bool b_force)
{
    assert(va);
    assert(cache);

    va->lock();

    const int num = vlc_array_count(&cache->subpictures);
    for (int i = num - 1; i >= 0; i--)
    {
        va_spu_cache_t *subpic;
        subpic = (va_spu_cache_t *)vlc_array_item_at_index(&cache->subpictures, i);
        if (!subpic)
            continue;

        vlc_array_remove(&cache->subpictures, i);

        assert(subpic->i_id != VA_INVALID_ID);
        assert(subpic->image->image_id != VA_INVALID_ID);

        vaDestroySubpicture(va->p_display, subpic->i_id);
        vaDestroyImage(va->p_display, subpic->image->image_id);

        subpic->i_id = VA_INVALID_ID;
        subpicture_cache_destroy(subpic);
    }
    if (b_force)
        cache->i_refcount = 0;
    cache_destroy(cache);
    cache = NULL;

    va->unlock();
}

static void va_spu_cache_ReleaseUnlocked(vlc_va_conn_t *va, vlc_va_spu_cache_t *spu, const int cache_id)
{
    assert(spu);

    spu_cache_t *cache = cache_find(&spu->cache, cache_id);
    if (cache)
    {
        cache_unref(cache);

        /* Do we still need need this subpicture cache ? */
        if (cache->i_refcount == 0)
        {
            int index = vlc_array_index_of_item(&spu->cache, (void*)cache);
            vlc_array_remove(&spu->cache, index);
            va_spu_cache_Destroy(va, cache, false);
        }
    }
}

/* External API */
int vlc_va_spu_cache_Link(vlc_object_t *obj, vlc_va_conn_t *va, vlc_va_spu_cache_t *spu,
                          const picture_t *picture, const int flags)
{
    assert(spu);
    assert(picture);
    assert(picture->p_sys);

    vlc_mutex_lock(&spu->lock);

    picture_sys_t *surface = picture->p_sys;
    assert(surface->i_id != VA_INVALID_SURFACE);

    spu_cache_t *cache = cache_find(&spu->cache, surface->i_cache);
    if (cache == NULL)
        goto error;

    assert(cache->i_id > 0);

    const int count = vlc_array_count(&cache->subpictures);
#ifdef VAAPI_DEBUG
    msg_Dbg(obj, "  link surface %d cache %d refcount %d",
                surface->i_id, cache->i_id, cache->i_refcount);
#endif

    int i_sub = 0;
    for (int i = 0; i < count; i++)
    {
        va_spu_cache_t *vasubpic;
        vasubpic = (va_spu_cache_t *) vlc_array_item_at_index(&cache->subpictures, i);
        if (!vasubpic)
            goto cleanup;

#ifdef VAAPI_DEBUG
        msg_Dbg(obj, "+ %d: subid %d", i_sub, vasubpic->i_id);
#endif
        /* NOTE: The video output core renders subpictures to picture size.
         * in case of VAAPI SURFACES this is always the original picture size.
         * Subpictures must be placed on the VAAPI SURFACE and vaPutSurface()
         * does the final scaling.
         */

        /* subpicture positioning */
        unsigned int src_x = 0;
        unsigned int src_w = src_x + vasubpic->rect.i_width;
        unsigned int src_y = 0;
        unsigned int src_h = src_y + vasubpic->rect.i_height;

        va->lock();

        /* Associate subpicture with surface */
        VAStatus status;
        status = vaAssociateSubpicture(va->p_display, vasubpic->i_id,
                                       &surface->i_id, 1, /* array of surfaces */
                                       /* src rectangle */
                                       src_x, src_y, src_w, src_h,
                                       /* dest rectangle */
                                       vasubpic->rect.i_x, vasubpic->rect.i_y,
                                       vasubpic->rect.i_width,
                                       vasubpic->rect.i_height,
                                       flags);

        va->unlock();
        if (status != VA_STATUS_SUCCESS)
            goto cleanup;

        subpicture_cache_link(vasubpic, &surface->i_id);
        vasubpic = NULL;
        i_sub++;
    }
    vlc_mutex_unlock(&spu->lock);
    return VLC_SUCCESS;

cleanup:
    /* Cleanup linked subpictures */
#ifdef VAAPI_DEBUG
    msg_Err(obj, "unlink surface %d cache %d", surface->i_id, cache->i_id);
#endif
    for (int i = 0; i < i_sub; i++)
    {
        va_spu_cache_t *sub = vlc_array_item_at_index(&cache->subpictures, i);
        assert(sub);

        if (sub->i_id != VA_INVALID_ID)
        {
            VAStatus status = VA_STATUS_ERROR_UNKNOWN;

            va->lock();
            status = vaDeassociateSubpicture(va->p_display,
                                             sub->i_id, &surface->i_id, 1);
            va->unlock();

            if (status != VA_STATUS_SUCCESS)
                msg_Err(obj, "failed unlinking subid: %d (error %d)",
                        sub->i_id, status);

            subpicture_cache_unlink(sub, &surface->i_id);
        }
    }

error:
    vlc_mutex_unlock(&spu->lock);
    return VLC_EGENERIC;
}

int vlc_va_spu_cache_Unlink(vlc_object_t *obj, vlc_va_conn_t *va,
                            vlc_va_spu_cache_t *spu, const picture_t *picture)
{
    assert(spu);
    assert(va);

    vlc_mutex_lock(&spu->lock);
    picture_sys_t *surface = picture->p_sys;

    assert(surface);
    assert(surface->i_id != VA_INVALID_SURFACE);
    assert(surface->i_cache > 0);

    spu_cache_t *cache = cache_find(&spu->cache, surface->i_cache);
    if (cache == NULL)
    {
        vlc_mutex_unlock(&spu->lock);
        return VLC_EGENERIC;
    }
    assert(cache->i_id > 0);

    const int count = vlc_array_count(&cache->subpictures);
    for (int i = 0; i < count; i++)
    {
        va_spu_cache_t *sub = vlc_array_item_at_index(&cache->subpictures, i);
        assert(sub);

        VAStatus status = VA_STATUS_ERROR_UNKNOWN;
        if (sub->i_id != VA_INVALID_ID)
        {
            va->lock();
            status = vaDeassociateSubpicture(va->p_display,
                                             sub->i_id, &surface->i_id, 1);
            va->unlock();

            if (status != VA_STATUS_SUCCESS)
                msg_Err(obj, "failed unlinking subid: %d (error %d)",
                        sub->i_id, status);

            subpicture_cache_unlink(sub, &surface->i_id);
        }
    }
    vlc_mutex_unlock(&spu->lock);
    return VLC_SUCCESS;
}

void vlc_va_spu_cache_Release(vlc_va_conn_t *va, vlc_va_spu_cache_t *spu, const int cache_id)
{
    vlc_mutex_lock(&spu->lock);
    va_spu_cache_ReleaseUnlocked(va, spu, cache_id);
    vlc_mutex_unlock(&spu->lock);
}

void vlc_va_spu_cache_Hold(vlc_va_spu_cache_t *spu, const int cache_id)
{
    assert(spu);

    vlc_mutex_lock(&spu->lock);
    spu_cache_t *cache = cache_find(&spu->cache, cache_id);
    if (cache)
        cache_ref(cache);
    vlc_mutex_unlock(&spu->lock);
}

spu_cache_t *vlc_va_spu_cache_Prepare(vlc_object_t *obj, vlc_va_conn_t *va, VAImageFormat *fmt,
                                      picture_t *picture, subpicture_t *subpicture, int flags)
{
    picture_sys_t *surface = picture->p_sys;

    spu_cache_t *cache = cache_new();
    if (unlikely(!cache))
        return NULL;

    cache->i_start = subpicture->i_start;
    cache->i_stop  = subpicture->i_stop;

#ifdef VAAPI_DEBUG
    int i_sub = 0;
#endif
    subpicture_region_t *region;
    for (region = subpicture->p_region; region != NULL; region = region->p_next)
    {
        /* Skip non-RGBA regions for now */
        if (region->fmt.i_chroma != VLC_CODEC_RGBA)
            continue;

        assert(region->p_picture->format.i_width > 0);
        assert(region->p_picture->format.i_height > 0);

        /* Create Subpicture */
        va_spu_cache_t *vasub_cache;
        vasub_cache = va_spu_cache_Create(obj, va, fmt, subpicture, region, flags);
        if (vasub_cache == NULL)
            goto cleanup;

#ifdef VAAPI_DEBUG
        msg_Dbg(obj, "+ %d: subid %d", i_sub, vasub_cache->i_id);
        i_sub++;
#endif

        vlc_array_append(&cache->subpictures, (void *)vasub_cache);
        vasub_cache = NULL;
    }

    /* remember what was created, so it can be released in DisplayPicture() */
    cache_ref(cache);
    surface->i_cache = cache->i_id;
    return cache;

cleanup:
    msg_Err(obj, "failed creating subpicture cache %d for surface %d",
            cache->i_id, surface->i_id);
    va_spu_cache_Destroy(va, cache, false);
    return NULL;
}

void vlc_va_spu_cache_Put(vlc_va_spu_cache_t *spu, spu_cache_t *cache)
{
    assert(spu);
    assert(cache);

    vlc_mutex_lock(&spu->lock);
    vlc_array_append(&spu->cache, (void *)cache);
    vlc_mutex_unlock(&spu->lock);
}

int vlc_va_spu_cache_Get(vlc_object_t *obj, vlc_va_spu_cache_t *spu, const unsigned int cache_id,
                         picture_t *picture, subpicture_t *subpicture)
{
    assert(spu);
    assert(picture);
    assert(subpicture);

    vlc_mutex_lock(&spu->lock);
    picture_sys_t *surface = picture->p_sys;

    spu_cache_t *cache = cache_find(&spu->cache, cache_id);
    if (unlikely(cache == NULL))
    {
        msg_Warn(obj, "subpicture cache id %d not found", cache_id);
        vlc_mutex_unlock(&spu->lock);
        return VLC_EGENERIC;
    }

    assert(cache->i_id == cache_id);

#if 0
    if ((cache->i_start != subpicture->i_start) &&
        (cache->i_stop  != subpicture->i_stop))
        goto cleanup;
#endif
    /* Check if all subpictures region pictures match */
    const int count = vlc_array_count(&cache->subpictures);
    int i_sub = 0;
    subpicture_region_t *region;
    for (region = subpicture->p_region; region != NULL; region = region->p_next)
    {
        /* Skip non-RGBA regions */
        if (region->fmt.i_chroma != VLC_CODEC_RGBA)
            continue;

        /* Ooh: big booboo
         * va_spu_cache_Create() failed for some of the subpicture regions. */
        if (i_sub >= count)
#ifdef VAAPI_DEBUG
            abort();
#else
            goto cleanup;
#endif

        va_spu_cache_t *vasubpic;
        vasubpic = (va_spu_cache_t *) vlc_array_item_at_index(&cache->subpictures, i_sub);
        if (!vasubpic)
            goto cleanup;

        if ((vasubpic->image->width  != region->fmt.i_width) ||
            (vasubpic->image->height != region->fmt.i_height))
            goto cleanup;

#ifdef VAAPI_DEBUG
        msg_Dbg(obj, "+ %d: subid %d", i_sub, vasubpic->i_id);
#endif
        vasubpic = NULL;
        i_sub++;
    }

    /* FIXME: should not happen */
    if (i_sub < count)
        goto cleanup;

    assert(cache->i_id == cache_id);
    surface->i_cache = cache->i_id;
    cache_ref(cache);
    vlc_mutex_unlock(&spu->lock);
    return VLC_SUCCESS;

cleanup:
    /* Cannot reuse this cached subpicture */
    msg_Warn(obj, "cannot reuse cached subpicture %d for surface %d -- resetting",
             cache->i_id, surface->i_id);
    vlc_mutex_unlock(&spu->lock);
    return VLC_EGENERIC;
}

vlc_va_spu_cache_t *vlc_va_spu_cache_New(void)
{
    vlc_va_spu_cache_t *spu = malloc(sizeof(vlc_va_spu_cache_t));
    if (!spu)
        return NULL;

    vlc_array_init(&spu->cache);
    vlc_mutex_init(&spu->lock);
    return spu;
}

void vlc_va_spu_cache_Delete(vlc_va_conn_t *va, vlc_va_spu_cache_t *restrict spu)
{
    vlc_mutex_lock(&spu->lock);
    const int count = vlc_array_count(&spu->cache);
    for (int i = count - 1; i >= 0; i--)
    {
        spu_cache_t *spu_cache;
        spu_cache = (spu_cache_t*) vlc_array_item_at_index(&spu->cache, i);
        if (!spu_cache)
            continue;

        vlc_array_remove(&spu->cache, i);

        va_spu_cache_ReleaseUnlocked(va, spu, vlc_va_spu_cache_GetID(spu_cache));
        va_spu_cache_Destroy(va, spu_cache, true);
    }
    vlc_array_clear(&spu->cache);
    vlc_mutex_unlock(&spu->lock);

    vlc_mutex_destroy(&spu->lock);
    free(spu);
}
